//***************************************************************************
// Group VDR/GraphTFT
// File comthread.c
// Date 28.10.06 - Jrg Wendel
// This code is distributed under the terms and conditions of the
// GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
//--------------------------------------------------------------------------
// Class ComThread
//***************************************************************************

#include "common.h"
#include "comthread.h"
#include "display.h"

//***************************************************************************
// Object
//***************************************************************************

ComThread::ComThread(void* aDisplay)
{
   line = new TcpChannel();

   display = aDisplay;
   bufferSize = maxBuf;

   buffer = new char[bufferSize+1];
   header = new TcpChannel::Header;

   running = false;
   timeout = 5;
   port = defaultPort;
   *host = 0;
   pid = 0;
   jpgQuality = 60;
}

ComThread::~ComThread()
{
   Stop();

   line->close();

   delete line;
   delete header;
   delete[] buffer;
}

//***************************************************************************
// Init
//***************************************************************************

int ComThread::init(Renderer* aRenderer, unsigned int aPort, const char* aHost)
{
   renderer = aRenderer;

   if (aHost)
      setHost(aHost);

   if (aPort)
      setPort(aPort);

   if (line->open(port) != 0)
   {
      d(1, "Error: Can't establish listener");
      return -1;
   }

   d(1, "Listener established!");
   
   Start();

   return 0;
}

//***************************************************************************
// Stop
//***************************************************************************

void ComThread::Stop() 
{
   if (running) 
   {
      running = false;
      Cancel(3);
   }
}

//***************************************************************************
// Run
//***************************************************************************

void ComThread::Action()
{
   int status;

   pid = getpid();
   isyslog("GraphTFT plugin tcp communication thread started (pid=%d)", pid);

   running = true;

   while (running)
   {
      if (!line->isConnected())
      {
         if (line->listen() == 0)
         {
            line->write(TcpChannel::cmdWelcome);
            d(1, "Client connection accepted");

            refresh();
         }

         sleep(1);

         continue;
      }

      while (line->isConnected() && running)
      {
         if ((status = line->look(10)) != 0)
         {
            if (status != TcpChannel::wrnNoEventPending)
            {
               d(1, "Error: Communication problems, closing line!");
               line->close();

               break;
            }
            
            continue;
         }
         
         if (read() != 0)
         {
            line->close();
            d(1, "Error: Communication problems, closing line!");
         }
      }
   }

   isyslog("GraphTFT plugin tcp communication thread ended (pid=%d)", pid);
}

//***************************************************************************
// Read
//***************************************************************************

int ComThread::read()
{
   int status;
   TcpChannel::Header tmp;

   // es stehen Daten, erst einmal den Header abholen ..

   if ((status = line->read((char*)&tmp, sizeof(TcpChannel::Header))) == 0)
   {
      header->command = ntohl(tmp.command);
      header->size = ntohl(tmp.size);

      switch (header->command)
      {
         case TcpChannel::cmdWelcome:
         {
            d(1, "Got welcome");
            break;
         }

         case TcpChannel::cmdLogout:
         {
            d(1, "Got logout from client, closing line");
            line->close();

            break;
         }

         case TcpChannel::cmdData:
         {
            d(1, "Got data");
            status = line->read(buffer, header->size);
            break;
         }

         case TcpChannel::cmdMouseEvent:
         {
            GraphTftTouchEvent ev;

            status = line->read((char*)&ev, header->size);
            d(1, "Got mouse event, button (%d) at (%d/%d)", ev.button, ev.x, ev.y); 
            ((cGraphTFTDisplay*)display)->mouseEvent(ev.x, ev.y, ev.button, ev.flag);

            break;
         }

         case TcpChannel::cmdStartCalibration:
         {
            ((cGraphTFTDisplay*)display)->setCalibrate(true);

            break;
         }

         case TcpChannel::cmdStopCalibration:
         {
            ((cGraphTFTDisplay*)display)->setCalibrate(false);

            break;
         }

         default:
         {
            d(eloAlways, "Got unexpected protocol (%d), aborting", header->command); 
            status = -1;

            break;
         }
      }
   }
   
   return status;
}

//***************************************************************************
// Refresh
//***************************************************************************

int ComThread::refresh()
{
   int size;

   if (line->isConnected())
   {

#ifdef HAVE_IMAGE_MAGICK

      const char* p = 0;

      if ((size = renderer->getImageBuffer((const void*&)p, jpgQuality)) > 0)
      {
         d(2, "Info: %ld kb to write", size/1024);
         
         if (line->write(TcpChannel::cmdData, p, size) != 0)
         {
            d(1, "Error: Communication problems, closing client connection.");
            line->close();
         }
      }

#else

      int fd;

      renderer->dumpImage2File("tmp.jpg", 720, 576);
      
      if ((fd = ::open("/tmp/tmp.jpg", O_RDONLY)) < 0)
      {
         d(1, "Error: Can't open dumped image!");
         return -1;
      }

      size = ::read(fd, buffer, maxBuf);
      
      if (size <= 0 || size == maxBuf)
      {
         d(1, "Error: Reading of image failed, result was %d."\
           "%s", size, size == maxBuf ? " buffer to smal" : strerror(errno));

         return -1;
      }
      
      d(2, "Info: %ld kb to write", size/1024);
      
      if (line->write(TcpChannel::cmdData, buffer, size) != 0)
      {
         d(1, "Error: Communication problems, closing client connection.");
         line->close();
      }

      ::close(fd);

#endif

   }

   return 0;
}
