/**
 *  GraphTFT plugin for the Video Disk Recorder 
 * 
 *  display.c - A plugin for the Video Disk Recorder
 *
 *  (c) 2004-2006 Lars Tegeler, Sascha Volkenandt, Jrg Wendel
 *
 * This code is distributed under the terms and conditions of the
 * GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
 *
 * $Id: display.c,v 1.2 2006/11/06 10:29:23 root Exp $
 * 
 **/

//***************************************************************************
// Includes
//***************************************************************************

#include <algorithm>
#include <sstream>

#include <vdr/menu.h>
#include <vdr/eitscan.h>
#include <vdr/remote.h>

#include "display.h"
#include "theme.h"
#include "setup.h"

// the renderdevices

#ifdef HAVE_IMLIB 
# include "fbrenderer.h"
# include "dvbrenderer.h"
# include "dmyrenderer.h"
#endif

#ifdef HAVE_DFB
# include "dfbrenderer.h"
#endif

#define SECONDS(x)  ((x)*1000)
#define MSECONDS(x) ((x)*1)

//***************************************************************************
// 
//***************************************************************************

const char* cGraphTFTService::normalModes[] =
{
   "Standard",
   "Clock",
   "Provider logo",
   "Sysinfo",

   0,
};

//***************************************************************************
// Object
//***************************************************************************

cGraphTFTDisplay::cGraphTFTDisplay()  
{
   // init renderer

   _render = 0;
   
   // init thread stuff

   _active = false;
   _marquee_active = false;
   _marquee_left = false;
   _marquee_idx = 0;
   _marquee_count = 0;
   _marquee_strip = 0;
   _updateIn = SECONDS(60);
   _recordingIn = 0;
   _volumeIn = 0;
   _messageIn = 0;
   _delayIn = 0;
   comThread = 0;

   // init animation

   _animation = false;
   _displayFrame = 0;
   _frameIn = 0;

   // init contents

   SetTabs(0);
   charInTabs[0] = charInTabs[1] = charInTabs[2] = charInTabs[3] = charInTabs[4] = charInTabs[5] = charInTabs[6] =  0;
   _mode = NormalView;
   channelType = ctTv;
   normalMode = nmStandard;
   _channel = 0;
   _presentTime = 0;
   _followingTime = 0;
   _volume = cDevice::CurrentVolume();
   _mute = cDevice::PrimaryDevice()->IsMute();
   _showRecording = 0;
   _replay.control = 0;
   _replay.lastMode = ModeUnknown;
   _menu.currentPosLast = 0;
   _menu.currentPos = 0;
   _menu.topPos = -1;
   _menu.lineHeight = -1;
   menuCount = 0;
   _Dumped = timeMs.Now();

   // calibration stuff

   calibrationPos.x = 100;
   calibrationPos.y = 100;

   // init pbp stuff

   _pbpActive = false;
}

cGraphTFTDisplay::~cGraphTFTDisplay() 
{
   Stop();

   if (comThread) delete comThread;
   if (_render) delete _render;
}

//***************************************************************************
// Init
//***************************************************************************

int cGraphTFTDisplay::Init(const char* dev, const char* cfgDir, int port)
{
   char* pos = 0;
   int devnum = -1;

#ifdef HAVE_DFB

   if ((pos = strstr(dev, "directFB")))
   {
      devnum = 0;
      _render = new DFbRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                720, 576,
                                GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf, 
                                theTheme->getDir());
   }

#endif
   
#ifdef HAVE_IMLIB

   if (!_render)
   {
      if ((pos = strstr(dev, "/dev/fb")))
      {
         devnum = atoi(pos+7);
         _render = new FbRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                  720, 576,
                                  GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf,
                                  theTheme->getDir());
      }
      
      else if ((pos = strstr(dev, "vdr/")))
      {    
         devnum = atoi(pos+4);
         _render = new DvbRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                   720, 576,
                                   GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf,
                                   theTheme->getDir());
      }
      
      else
      {
         _render = new DummyRenderer(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                                     720, 576,
                                     GraphTFTSetup.PluginConfPath, GraphTFTSetup.Iso2Utf,
                                     theTheme->getDir());
      }
      
      comThread = new ComThread(this);
      
      if (comThread->init(_render, port) != 0)
      {
         d(1, "Cant establish listener, tcp communication not available");
         delete comThread;
         comThread = 0;
      }
   }
   
#endif

   if (!comThread && devnum == -1) 
      return 0;

   _render->setBorder(GraphTFTSetup.VdrWidth, GraphTFTSetup.VdrHeight);

   if (_render->init(devnum) != 0)
      return 0;

   _render->clear();

   // Show fist black Background

   _render->image(theTheme->getStartImage().c_str(), 0,0,0,0);
   _render->refresh();
   _render->setPlayMode(false);

   Start();

   return 1;
}

//***************************************************************************
// Setup Changed
//***************************************************************************

void cGraphTFTDisplay::setupChanged()
{
   logDevice = GraphTFTSetup.LogDevice;
   logLevel = GraphTFTSetup.Level;

   comThread->setJpegQuality(GraphTFTSetup.JpegQuality);

   if (_render)
   {
      _render->setBorder(GraphTFTSetup.VdrWidth, GraphTFTSetup.VdrHeight);
      _render->setProperties(GraphTFTSetup.VdrX, GraphTFTSetup.VdrY,
                             720, 576,
                             GraphTFTSetup.Iso2Utf,
                             theTheme->getDir());
   }
}

//***************************************************************************
// Set Calibrate
//***************************************************************************

void cGraphTFTDisplay::setCalibrate(int active)
{
   d(1, "Info: %s calibration mode", active ? "starting" : "stopping");

   if (active)
      setMode(cGraphTFTDisplay::Calibration);
   else
      setMode(cGraphTFTDisplay::NormalView, /*force*/ true);

   _doUpdate.Broadcast();
}

//***************************************************************************
// Stop
//***************************************************************************

void cGraphTFTDisplay::Stop() 
{
   if (_active) 
   {
      _active = false;
      _doUpdate.Broadcast();
      Cancel(3);
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::Pbp() 
{
   d(1, "Pbp()");

   if (!_pbpActive)
   {
      PbpInit();
      PbpStart();
   }
   else
   {
      PbpStop();
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpRestart()
{
   d(1, "PbpRestart()");

   if (_pbpActive)
   {
      PbpStop();
      PbpStart();
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpSwitch(bool Back) 
{
   const cChannel* newChan = 0;

   d(1, "PbpSwitch()");

   if (_pbpActive)
   {
      PbpStop();

      if (!Back)
      {
         newChan = cDevice::CurrentChannel() != 0 ? Channels.GetByNumber(cDevice::CurrentChannel()) : NULL;
      }

      cDevice::PrimaryDevice()->SwitchChannel(_chan, true);

      if ( !Back )
      {
         if (newChan != NULL)
         {
            _chan = newChan;
            _dev = cDevice::GetDevice(_chan, 1);
            PbpStart();
         }
      }
   }
}


//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpUpDown(bool Up) 
{
   d(1, "display.c: PbpUpDown()");

   if (_pbpActive)
   {
      // Set new Channel and device
      // but not today ;-)

      if (Up)
      {
         // set TFT Channel one up
      }
      else
      {
         // and here down
      }

      // Now make a restart

      // PbpRestart();
   }
}

//***************************************************************************
// 
//***************************************************************************

bool cGraphTFTDisplay::PbpIsRunning()
{
   d(1, "PbpIsRunning() (%s)", _pbpActive ? "yes" : "no");

   return _pbpActive;
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpInit() 
{
   d(1, "PbpInit()");

   _chan = cDevice::CurrentChannel() != 0 ? Channels.GetByNumber(cDevice::CurrentChannel()) : 0;

   if (_chan)
      _dev = cDevice::GetDevice(_chan, 1);

      //_dev = cDevice::ActualDevice();
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpStart() 
{
   d(1, "PbpStart() dev '%s', chan '%s'", 
     _dev  ? "yes" : "no", 
     _chan ? "yes" : "no");

   if (_dev && _chan)
   {
      d(1 , "Set PlayMode to Video\n");

      Stop();
      
      _render->setPlayMode(true);
      _dev->SwitchChannel(_chan, false);

      // active transfer

      _transfer = new cGraphTFTTransfer(_render, _chan);
      _dev->AttachReceiver(_transfer);

      _pbpActive = true;
   }
}

//***************************************************************************
// 
//***************************************************************************

void cGraphTFTDisplay::PbpStop() 
{
   d(1, "PbpStop()");

   if (_transfer) delete _transfer;

   _render->setPlayMode(false);
   _pbpActive = false;
   Start();
   _doUpdate.Broadcast();
}

//***************************************************************************
// Set Tabs
//***************************************************************************

void cGraphTFTDisplay::SetTabs(int Tab1, int Tab2, int Tab3, int Tab4, int Tab5)
{
   d(6, "SetTabs - Tab1: %d, Tab2, %d, Tab3 %d, Tab4 %d, Tab5 %d",Tab1, Tab2, Tab3, Tab4, Tab5);
   
   tabs[0] = 0;
   tabs[1] = Tab1 ? tabs[0] + Tab1 : 0;
   tabs[2] = Tab2 ? tabs[1] + Tab2 : 0;
   tabs[3] = Tab3 ? tabs[2] + Tab3 : 0;
   tabs[4] = Tab4 ? tabs[3] + Tab4 : 0;
   tabs[5] = Tab5 ? tabs[4] + Tab5 : 0;
   tabs[6] = 100;
}

//***************************************************************************
// Get Tabbed Text
//***************************************************************************

const char* cGraphTFTDisplay::GetTabbedText(const char *s, int Tab)
{
   if (!s)
      return 0;

   static char buffer[1000];
   const char *a = s;
   const char *b = strchrnul(a, '\t');

   while (*b && Tab-- > 0) 
   {
      a = b + 1;
      b = strchrnul(a, '\t');
   }

   if (!*b)
      return (Tab <= 0) ? a : 0;

   unsigned int n = b - a;

   if (n >= sizeof(buffer))
      n = sizeof(buffer) - 1;

   strncpy(buffer, a, n);
   buffer[n] = 0;

   return buffer;
}

//***************************************************************************
// Action
//***************************************************************************

void cGraphTFTDisplay::Action()
{
   int wait = 0, elapsed;

   isyslog("GraphTFT plugin display thread started (pid=%d)", getpid());

   // give the plugin time to collect some data AND give vdr some time to
   // finish initialization before acquiring the lock

   usleep(500000);

   _active = true;
   _mutex.Lock(); // mutex gets ONLY unlocked when sleeping

   while (_active)
   {
      d(1, "action loop");

      // milliseconds elapsed since last update

      elapsed = wait ? timeMs.Elapsed() - wait : 0;

      timeMs.Set();

      // check for volumebar

      if (_volumeIn < 0)
         _volumeIn = 0;         // _volumeIn *= -1;
      else if (_volumeIn > 0)
         _volumeIn -= elapsed;

      // check for osdmessage 

      if (_messageIn < 0)
         _messageIn = 0;
      else if (_messageIn > 0)
         _messageIn -= elapsed;

      // check for animation

      if (_animation && _frameIn > 0)
         _frameIn -= elapsed;

      // check if there is another recording to show

      if (_recordings.size() > 1) 
      {
         if (_recordingIn < 0)
            _recordingIn *= -1;
         else 
         {
            _recordingIn -= elapsed;

            if (_recordingIn <= 0) 
            {
               _showRecording = (_showRecording + 1) % _recordings.size();
               _recordingIn = SECONDS(5);
            }
         }
      }

      // here do the update

      wait = timeMs.Elapsed();

      if (isModeNormal(_mode))
      {
         switch (normalMode)
         {
            case nmStandard: Display(channelType == ctTv ? sectionNormalTV : sectionNormalRadio); break;
            case nmClock:    Display(sectionNormalClock);  break;
            case nmLogo:     Display(sectionNormalLogo);   break;
            case nmSoccer:   Display(sectionNormalSoccer); break;
               
            default:         Display(sectionNormalTV);
         }
      }
      else
      {
         switch (_mode) 
         {
            case ReplayNormal:      Display(sectionReplayNormal);    break;
            case ReplayMP3:         Display(sectionReplayMP3);       break;
            case ReplayDVD:         Display(sectionReplayDVD);       break;

            case MenuWhatsOnElse:   Display(sectionMenuWhatsOnElse); break;
            case MenuWhatsOnNow:    Display(sectionMenuWhatsOnNow);  break;
            case MenuWhatsOnNext:   Display(sectionMenuWhatsOnNext); break;
            case MenuMain:          Display(sectionMenuMain);        break;
            case MenuSchedule:      Display(sectionMenuSchedule);    break;
            case MenuChannels:      Display(sectionMenuChannels);    break;
            case MenuTimers:        Display(sectionMenuTimers);      break;
            case MenuRecordings:    Display(sectionMenuRecordings);  break;
            case MenuSetup:         Display(sectionMenuSetup);       break;
            case MenuSetupPage:     Display(sectionMenuSetupPage);   break;
            case MenuCommands:      Display(sectionMenuCommands);    break;
            case MenuDefault:       Display(sectionMenu);            break;

            case Calibration:       Display(sectionCalibration);     break;

            default:         Display(sectionNormalTV);
         }
      }

      elapsed = timeMs.Elapsed() - wait;
      
      // next update in...?

      if (isModeReplay(_mode))
         _updateIn = SECONDS(1) - elapsed;      // one second minus update time
      else if (_recordings.size() > 1)
         _updateIn = _recordingIn;              // use time calculated above
      else
         _updateIn = SECONDS(60);               // if nothing changed

      if (!_marquee_active)
      {
         _updateIn = SECONDS(60); // one minute
         _marquee_active = true;
      }

      // possibly updating osdmessage 

      if (_messageIn > 0 && _messageIn < _updateIn)
         _updateIn = _messageIn;
      
      // possibly updating volume bar

      if (_volumeIn > 0 && _volumeIn < _updateIn)
         _updateIn = _volumeIn;

      // possibly animating

      //       if (_animation && !isModeMenu(_mode))
      //       {
      //          _frameIn -= elapsed;

      //          if (_frameIn < 0)
      //             _updateIn = 0;
      //          else if (_frameIn < _updateIn)
      //             _updateIn = _frameIn;
      //       }

      // wait for event or timeout

      wait = timeMs.Elapsed();

      _doUpdate.TimedWait(_mutex, _updateIn);

      while (_delayIn)
      {
         // update delayed was requested

         d(5, "delayed update in max. %d ms\n", _delayIn);
         _doUpdate.TimedWait(_mutex, _delayIn);
         _delayIn = 0;
      }
   }

   isyslog("GraphTFT plugin display thread ended (pid=%d)", getpid());
}

//***************************************************************************
// Stopped Timer
//***************************************************************************

bool cGraphTFTDisplay::StoppedTimer(const char *Name) 
{
   cTimer* timer = Timers.First();

   while (timer)
   {
      if (strcmp(Name, timer->File()) == 0)
         break;

      timer = Timers.Next(timer);
   }

   return timer == 0 || !timer->Recording();
}

//***************************************************************************
// Update Programme
//***************************************************************************

void cGraphTFTDisplay::UpdateProgramme()
{
   const cChannel *channel = Channels.GetByNumber(_channel);

   if (channel) 
   {
      cMutexLock lock(&_mutex);

      const cEvent *present = 0, *following = 0;
      cSchedulesLock schedulesLock;
      const cSchedules *schedules = cSchedules::Schedules(schedulesLock);

      if (schedules) 
      {
         const cSchedule *schedule = schedules->GetSchedule(channel->GetChannelID());

         if (schedule) 
         {
            if ((present = schedule->GetPresentEvent()) != 0) 
            {
               _presentTime = present->StartTime();
               _presentTitle = present->Title() ? present->Title() : "";
               _presentSubtitle = present->ShortText() ? present->ShortText() : "";
            }

            if ((following = schedule->GetFollowingEvent()) != 0) 
            {
               _followingTime = following->StartTime();
               _followingTitle = following->Title() ? following->Title() : "";
               _followingSubtitle = following->ShortText() ? following->ShortText() : "";
            }
         }
      }
   }
}

//***************************************************************************
// Osd Channel Switch
//***************************************************************************

void cGraphTFTDisplay::ChannelSwitch(const cDevice *Device, int ChannelNumber) 
{
   d(3, "ChannelSwitch on %p: %d\n", Device, ChannelNumber);

   if (Device->IsPrimaryDevice() && !EITScanner.UsesDevice(Device) && _channel != ChannelNumber) 
   {
      const cChannel* channel;

      cMutexLock lock(&_mutex);
      _channel = ChannelNumber;

      if ((channel = Channels.GetByNumber(_channel)) != 0)
      {
         char *temp;
         int pos;

         asprintf(&temp, "%d %s", channel->Number(), channel->Name());
         _channelText = temp;

         if ((pos = _channelText.rfind(';')) != -1)
            _channelText.erase(pos);
         if ((pos = _channelText.rfind(',')) != -1)
            _channelText.erase(pos);

         free(temp);

         if (!isModeMenu(_mode)) 
         {
            switch (channel->Vpid()) 
            {
               case 0:
               case 1:
               case 0x1fff:
                  channelType = ctRadio;
                  break;
                  
               default:
                  channelType = ctTv;
                  break;
            }

            setMode(NormalView);
         }
      }

      _delayIn = 100;
      _marquee_active = false;

      if (GraphTFTSetup.ViewType == 2 || GraphTFTSetup.ViewType == 3)
         _updateIn = MSECONDS(GraphTFTSetup.ScrollTime); 

      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Channel
//***************************************************************************

void cGraphTFTDisplay::OsdChannel(const char *Text) 
{
   d(5, "OsdChannel: %s\n", Text);

   if (_channelInfo != Text)
   {
      int pos;
      string temp = Text;
      cMutexLock lock(&_mutex);

      // workaround for some osd themes (Elchi)

      _channelInfo = "";
      _presentTitle = "";
      _presentSubtitle = "";
      _followingTitle = "";
      _followingSubtitle = "";

      if ((pos = temp.find(' ')) != -1) 
      {
         _channelInfo += temp.substr(0, pos) + " ";

         while (pos < (int)temp.length() && temp[pos] == ' ')
            ++pos;

         int l = temp.length();

         while (temp[l - 1] == ' ')
            --l;

         _channelInfo += temp.substr(pos, l - pos);
      } 
      else
         _channelInfo = Text;

      if ((pos = _channelInfo.rfind(';')) != -1)
         _channelInfo.erase(pos);

      if ((pos = _channelInfo.rfind(',')) != -1)
         _channelInfo.erase(pos);

      _delayIn = 100;
      _marquee_active = false;

      if (GraphTFTSetup.ViewType == 2 || GraphTFTSetup.ViewType == 3) 
         _updateIn = MSECONDS(GraphTFTSetup.ScrollTime); 

      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Programme
//***************************************************************************

void cGraphTFTDisplay::OsdProgramme(time_t PresentTime, const char *PresentTitle, const char *PresentSubtitle, time_t FollowingTime, const char * FollowingTitle, const char * FollowingSubtitle)
{
   d(5, "OsdProgramme at %u: (Present)   %s, %s\n", (uint)PresentTime, PresentTitle, PresentSubtitle);
   d(5, "OsdProgramme at %u: (Following) %s, %s\n", (uint)FollowingTime, FollowingTitle, FollowingSubtitle);

   // check all pointers before (STL doesn't like comparing against NULL)

   if (!PresentTitle)       PresentTitle = "";
   if (!PresentSubtitle)    PresentSubtitle = "";
   if (!FollowingTitle)     FollowingTitle = "";
   if (!FollowingSubtitle)  FollowingSubtitle = "";

   if (_presentTime != PresentTime || _presentTitle != PresentTitle || _presentSubtitle != PresentSubtitle || _followingTime != FollowingTime || _followingTitle == FollowingTitle || _followingSubtitle != FollowingSubtitle) 
   {
      cMutexLock lock(&_mutex);
      _presentTime = PresentTime;
      _presentTitle = PresentTitle ? PresentTitle : "";
      _presentSubtitle = PresentSubtitle ? PresentSubtitle : "";
      _followingTime = FollowingTime;
      _followingTitle = FollowingTitle ? FollowingTitle : "";
      _followingSubtitle = FollowingSubtitle ? FollowingSubtitle : "";
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Set Volume
//***************************************************************************

void cGraphTFTDisplay::SetVolume(int, bool) 
{
   d(3, "SetVolume to %d, muted is %s\n", cDevice::CurrentVolume(), cDevice::PrimaryDevice()->IsMute() ? "true" : "false");

   if (_volume != cDevice::CurrentVolume() || _mute != cDevice::PrimaryDevice()->IsMute()) 
   {
      cMutexLock lock(&_mutex);
      _volume = cDevice::CurrentVolume();
      _mute = cDevice::PrimaryDevice()->IsMute();
      _volumeIn = SECONDS(3); // -SECONDS(2);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Recording
//***************************************************************************

void cGraphTFTDisplay::Recording(const cDevice *Device, const char *Name, const char *FileName, bool On) 
{
   d(3, "Recording %s to %p\n", Name, Device);

   if (On) 
   {
      if (!Name) Name = "";
      RecordingInfo info;
      cMutexLock lock(&_mutex);
      info.device = Device;
      info.name = Name;
      _recordings.push_back(info);
      _recordingIn = -SECONDS(5);
      _doUpdate.Broadcast();
   } 
   else 
   {
      cMutexLock lock(&_mutex);
      Recordings::iterator it = _recordings.begin();

      for (; it != _recordings.end(); ++it) 
      {
         if ((*it).device == Device && StoppedTimer((*it).name.c_str())) 
         {
            d(5, "deleting recording %s\n", (*it).name.c_str());
            _recordings.erase(it);
            break;
         }
      }

      if (_recordings.size() > 0) 
      {
         _showRecording = _showRecording % _recordings.size();
         _recordingIn = -SECONDS(5);
      }

      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Replaying
//***************************************************************************

void cGraphTFTDisplay::Replaying(const cControl *Control, const char *Name, 
                                 const char *FileName, bool On) 
{
   d(3, "Replaying %s '%s' from %p\n", On ? "start" : "stop", Name, Control);

   if (On) 
   {
      if (!Name) Name = "";

      cMutexLock lock(&_mutex);
      _replay.control = (cControl *)Control;
      _replay.lastMode = _mode;
      setMode(ReplayNormal);

      if (strlen(Name) > 6 && Name[0]=='[' && Name[3]==']' && Name[5]=='(') 
      {
         int i;

         for (i = 6; Name[i]; ++i) 
         {
            if (Name[i] == ' ' && Name[i-1] == ')')
               break;
         }

         if (Name[i]) 
         { 
            // replaying mp3

            const char* name = skipspace(Name + i);
            _replay.name = *name ? name : tr("Unknown title");
            setMode(ReplayMP3);
         }
      } 
      else if (strcmp(Name, "DVD") == 0) 
      {
         _replay.name = Name;
         setMode(ReplayDVD);
      } 
      else if (strlen(Name) > 7) 
      {
         int i, n;

         for (i = 0, n = 0; Name[i]; ++i) 
         {
            if (Name[i] == ' ' && Name[i-1] == ',' && ++n == 4)
               break;
         }

         if (Name[i]) 
         { 
            // replaying DVD

            const char *name = skipspace(Name + i);
            _replay.name = *name ? name : tr("Unknown title");
            replace(_replay.name.begin(), _replay.name.end(), '_', ' ');
            setMode(ReplayDVD);
         }
      }

      if (_mode == ReplayNormal) 
      {
         int pos;
         _replay.name = Name;
         _replay.folder = Name;

         if ((pos = _replay.name.rfind('~')) != -1) 
         {
            _replay.name.erase(0, pos + 1);
            _replay.folder.erase(pos);
         }
      }

      _doUpdate.Broadcast();
   } 

   else 
   {
      cMutexLock lock(&_mutex);
      _replay.control = 0;
      _replay.name = "";
      _replay.folder = "";
      setMode(NormalView);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Status Message
//***************************************************************************

void cGraphTFTDisplay::OsdStatusMessage(const char *Message) 
{
   d(5, "OsdStatusMessage %s\n", Message);

   if (!Message) 
   {
      cMutexLock lock(&_mutex);
      _message = "";
      _doUpdate.Broadcast();
   } 
   else if (_message != Message) 
   {
      cMutexLock lock(&_mutex);
      _message = Message;
      _messageIn = SECONDS(GraphTFTSetup.OSDMessageTime);
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Title
//***************************************************************************

void cGraphTFTDisplay::OsdTitle(const char* Title) 
{
   d(5, "OsdTitle %s\n", Title);

   if (Title) 
   {
      int pos;
      cMutexLock lock(&_mutex);

      _message = "";
      _menu.title = Title;

      if ((pos = _menu.title.find('\t')) != -1)
         _menu.title.erase(pos);

      if (_mode == MenuWhatsOnElse && (strstr(Title, tr("What's on now?"))))
      {
         d(1, "Assume mode 'whats on now' due to menue title '%s'", Title);
         setMode(MenuWhatsOnNow);
      }
   }
}

//***************************************************************************
// Osd Item
//***************************************************************************

void cGraphTFTDisplay::OsdItem(const char* Text, int Index) 
{
   d(5, "OsdItem %s at %d\n", Text, Index);

   if (Text)
   {
      cMutexLock lock(&_mutex);
      MenuItem item;
      item.text = Text;
      item.type = itNormal;
      item.tabCount = 0;

      _message = "";

      for (int i = 0; i < MaxTabs; ++i) 
      {
         const char* tp = GetTabbedText(Text, i);

         if (tp)
         {
            char* tab;
            int len;

            tab = strdup(tp);
            strreplace(tab, '\177', '|');
            strreplace(tab, '\203', ' ');
            strreplace(tab, '\204', ' ');
            strreplace(tab, ']', ' ');
            strreplace(tab, '[', ' ');

            len = strlen(Str::allTrim(tab));

            if (charInTabs[i] < len)
               charInTabs[i] = len;

            // detect partingLine
            // some special handling for lines of epgsearch plugin like
            // ---------------------------------- \t Die 24.10.2006 ------------------------

            if (i == 1 && strstr(item.tabs[0].c_str(), "-----"))
            {
               item.type = itPartingLine;
               char* p;

               if ((p = strstr(tab, " ---")))
                  *p = 0;

               if (!strstr(tab, "-----"))
                  item.tabs[0] = Str::allTrim(tab);
               else
                  item.tabs[0] = "";

               d(2, "Detected parting line '%s'", item.tabs[0].c_str());
            }
            else
            {
               item.tabs[i] = tab;
            }

            item.tabCount++;
            free(tab);
         }
         else
            break;
      }
      
      _menu.items.push_back(item);
   }
}

//***************************************************************************
// Osd Current Item
//***************************************************************************

void cGraphTFTDisplay::OsdCurrentItem(const char *Text)
{
   d(3, "OsdCurrentItem %s\n", Text);

   if (Text) 
   {
      cMutexLock lock(&_mutex);
      int i;
      _message = "";
      _menu.current = Text;

      for (i = 0; i < (int)_menu.items.size(); i++) 
      {
         if (_menu.items[i].text == _menu.current)
         {
            _menu.currentPos = i;
            break;
         }
      }

      // If we found no current Pos it may be the old one with changed values

      if (_menu.currentPos < (int)_menu.items.size() 
          && i == (int)_menu.items.size() 
          && strcmp(_menu.items[_menu.currentPos].tabs[0].c_str(), GetTabbedText(Text, 0)) == 0)
      {
         d(5, "OsdCurrentItem %s", Text);

#if __GNUC__ < 3
         MenuItem item = _menu.items[_menu.currentPos];
#else
         MenuItem item = _menu.items.at(_menu.currentPos);
#endif

         vector<MenuItem>::iterator tempIterator;
         tempIterator = _menu.items.begin();

         for (int i = 0; i < _menu.currentPos; i++)
            tempIterator++;

         tempIterator = _menu.items.erase(tempIterator);
         item.text = Text;

         for (int i = 0; i < MaxTabs; ++i) 
         {
            const char *tab = GetTabbedText(Text, i);

            if (tab)
               item.tabs[i] = tab;

            if (GetTabbedText(Text, i+1) == 0)
               break;
         }

         _menu.items.insert(tempIterator, item);
         d(5, "OsdCurrentItem %s added\n", Text);
      }      

      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Clear
//***************************************************************************

void cGraphTFTDisplay::OsdClear(void) 
{
   d(3, "OsdClear\n");

   cMutexLock lock(&_mutex);
   _message = "";
   _menu.title = "";
   _menu.items.clear();
   charInTabs[0] = charInTabs[1] = charInTabs[2] = charInTabs[3] = charInTabs[4] = charInTabs[5] = 0;
   SetTabs(0,0,0,0,0);

   for (int i = 0; i < 4; ++i)
      _menu.buttons[i] = "";

   _menu.text = "";

   if (isModeMenu(_mode) || _channelInfo != "") 
   {
      _channelInfo = "";
      _delayIn = 100;
      _doUpdate.Broadcast();
   }
}

//***************************************************************************
// Osd Help Keys
//***************************************************************************

void cGraphTFTDisplay::OsdHelpKeys(const char *Red, const char *Green, const char *Yellow, const char *Blue) 
{
   d(5, "OsdHelpKeys: %s, %s, %s, %s\n", Red, Green, Yellow, Blue);

   cMutexLock lock(&_mutex);
   _menu.buttons[0] = Red ? Red : "";
   _menu.buttons[1] = Green ? Green : "";
   _menu.buttons[2] = Yellow ? Yellow : "";
   _menu.buttons[3] = Blue ? Blue : "";
}

//***************************************************************************
// Osd Text Item
//***************************************************************************

void cGraphTFTDisplay::OsdTextItem(const char *Text, bool Scroll) 
{
   d(5, "OsdTextItem: %s scroll %s\n", Text, Scroll ? "true" : "false");

   cMutexLock lock(&_mutex);

   if (Text)
   { 
      _menu.text = Text;
      _menu.textScroll = 0;
      replace(_menu.text.begin(), _menu.text.end(), '\n', ' ');
   }
   else
   {
      if (Scroll)
         _menu.textScroll--;
      else
         _menu.textScroll++;
   }

   _doUpdate.Broadcast();
}

//***************************************************************************
// Osd Menu Display
//***************************************************************************

void cGraphTFTDisplay::OsdMenuDisplay(eOsdMenuKind kind)
{
   d(3, "OsdMenuDisplay - kind %d", kind);

   cMutexLock lock(&_mutex);

   if (!isModeMenu(_mode)) 
      _menu.lastMode = _mode;
   
   switch (kind)
   {
      case omMain:            setMode(MenuMain);        break;
      case omSetup:           setMode(MenuSetup);       break;
      case omChannels:        setMode(MenuChannels);    break;
      case omCommands:        setMode(MenuCommands);    break;
      case omTimers:          setMode(MenuTimers);      break;
      case omSchedule:        setMode(MenuSchedule);    break; // program
      case omRecordings:      setMode(MenuRecordings);  break;
      case omWhatsOnNow:      setMode(MenuWhatsOnNow);  break; // jetzt
      case omWhatsOnElse:     setMode(MenuWhatsOnElse); break; // abend, nacht, ..
      case omWhatsOnNext:     setMode(MenuWhatsOnNext); break; // next
      case omSetupPage:       setMode(MenuSetupPage);   break;

      case omText: 
      case omTimerEdit:
      case omEvent:
      case omCam:
      case omCamEnquiry:
      case omEditChannel:

      case omRecording:
      case omRenameRecording: setMode(MenuDefault);    break;

      default:                setMode(MenuDefault);    break;
   }
}

//***************************************************************************
// Osd Menu Destroy
//***************************************************************************

void cGraphTFTDisplay::OsdMenuDestroy()
{
   d(3, "OsdMenuDestroy");

   cMutexLock lock(&_mutex);

   // all menues closed !

   if (isModeMenu(_mode))        // 'if' is needed for replay via extrecmenu
      setMode(_menu.lastMode);      
}

//***************************************************************************
// Mouse Event
//***************************************************************************

void cGraphTFTDisplay::mouseEvent(int x, int y, int button, int flag)
{
   if (!currentSection)
      return ;

   cMutexLock lock(&_mutex);

   if (isModeCallibration(_mode))
   {
      calibrationPos.x = x;
      calibrationPos.y = y;

      _doUpdate.Broadcast();
   }

   else if (button == ComThread::mbWheelUp)
   {
      cRemote::Put(cKey::FromString("Up"));
   }

   else if (button == ComThread::mbWheelDown)
   {
      cRemote::Put(cKey::FromString("Down"));
   }

   else if (button == ComThread::mbRight)
   {
      cRemote::Put(cKey::FromString("Back"));
   }

   else if (button == ComThread::mbLeft)
   {
      for (cGraphTFTThemeItem* p = currentSection->getItems()->First(); p; 
           p = currentSection->getItems()->Next(p))
      {
         // check click coordinates 
         
         if (x >= p->X() && x <= p->X()+p->Width() &&
             y >= p->Y() && y <= p->Y()+p->Height())
         {
            d(2, "item %d on (%d/%d) clicked, sending key '%s', %sfound in keytab", p->Item(),
              x, y, p->OnClick().c_str() ? p->OnClick().c_str() : "<none>",
              p->OnClick().c_str() && cKey::FromString(p->OnClick().c_str()) ? "" : "not ");

            // menue navigation area ?
            
            if (p->Item() == itemMenuNavigationArea &&
                _menu.lineHeight > 0)
            {
               int yOff = y - p->Y();
               int clickRow = yOff / _menu.lineHeight + _menu.topPos;
               
               if (_menu.currentPos < clickRow)        // down
                  for (int i = _menu.currentPos; i < clickRow; i++)
                     cRemote::Put(cKey::FromString("Down"));
               else if (_menu.currentPos > clickRow)   // up
                  for (int i = _menu.currentPos; i > clickRow; i--)
                     cRemote::Put(cKey::FromString("Up"));
            }
            
            // check if item with key definition and double clicked
            
            if (p->OnClick() != "" && flag == 0)
            {
               d(2, "item %d clicked, sending key '%s'", p->Item(), 
                 p->OnClick().c_str());

               cRemote::Put(cKey::FromString(p->OnClick().c_str()));
            }

            if (p->OnDblClick() != "" && (flag & ComThread::efDoubleClick))
            {
               d(2, "item %d double-clicked, sending key '%s'", p->Item(), 
                 p->OnDblClick().c_str());
               
               cRemote::Put(cKey::FromString(p->OnDblClick().c_str()));
            }
         }
      }
   }
}

//***************************************************************************
// Display
//***************************************************************************

void cGraphTFTDisplay::Display(int sectionId)
{
   static int lastSectionId = sectionUnknown;

   cGraphTFTThemeItem* pSelect = 0;

   DisplayClear();

   if (isModeNormal(_mode))
      UpdateProgramme();

   else if (isModeMenu(_mode) && !theTheme->getSection(sectionId))
   {
      d(2, "Info: Section faked to '%s' due to section '%s' not defined!", 
        cThemeService::toName((eThemeSection)sectionMenu), cThemeService::toName((eThemeSection)sectionId));

      sectionId = sectionMenu;
   }

   if (sectionId != lastSectionId)
   {
      d(0, "Section changed from '%s' to '%s'", 
        cThemeService::toName((eThemeSection)lastSectionId),
        cThemeService::toName((eThemeSection)sectionId));

      lastSectionId = sectionId;
      _menu.topPos = -1;
      _menu.lineHeight = -1;
   }

   if (!(currentSection = theTheme->getSection(sectionId)))
      return ;

   for (cGraphTFTThemeItem* p = currentSection->getItems()->First(); p; p = currentSection->getItems()->Next(p))
   {
      switch (p->Item())
      {
         case itemImage:
            DisplayImage(p->X(), p->Y(), p->Width(),p->Height(),p->Transparent(),p->Path());
            break;

         case itemImageCover:
            DisplayImageCover(p->X(), p->Y(), p->Width(), p->Height(), p->Transparent(),p->Path());
            break;
               
         case itemImageFile:
            DisplayImageFile(p->X(), p->Y(), p->Width(), p->Height(), p->Transparent(), p->Path(), p->Path2());
            break;
               
         case itemText:
            DisplayText(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines(), p->Text(),false);
            break;
               
         case itemTimebar:
            DisplayTimebar(p);
            break;
               
         case itemDate:
            DisplayDate(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemDateOnly:
            DisplayDateOnly(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(),p->Align(), p->Lines());
            break;
               
         case itemTimeOnly:
            DisplayTimeOnly(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(),p->Align(), p->Lines());
            break;

         case itemDateTimeFormat:
            DisplayDateTimeFormat(p);
            break;

         case itemLogo:
            DisplayLogo(p->X(), p->Y(), p->Width(),p->Height(),p->Transparent(),p->Path(), p->Type(), p->Animated(), p->Delay(), p->Count());
            break;
               
         case itemSymRecording:
            DisplayRecordingSymbol(p->X(), p->Y(), p->Width(),p->Height(),p->Transparent(),p->Path(),p->Path2());
            break;
               
         case itemSym2ch:
         case itemSymDD:       
         case itemSymVTX:       
         case itemSymCrypt:
         case itemSymRepPlay:
         case itemSymRepPause:
         case itemSymRepFF:
         case itemSymRepFB:
         case itemSymRepSF:
         case itemSymRepSB:
            DisplaySymbols(p->X(), p->Y(), p->Width(), p->Height(), p->Transparent(),
                           p->Path(), p->Path2(), p->Item());
            break;
               
         case itemChannel:
            DisplayChannelName(p->X(), p->Y(), p->Width(), p->Height(),
                               p->Red(), p->Green(), p->Blue(), p->Transparent(), 
                               p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemChannelOnly:
            DisplayChannelNameOnly(p->X(), p->Y(), p->Width(), p->Height(),
                                   p->Red(), p->Green(), p->Blue(), p->Transparent(), 
                                   p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemChannelNumber:
            DisplayChannelNumber(p->X(), p->Y(), p->Width(), p->Height(), p->Red(), 
                                 p->Green(), p->Blue(), p->Transparent(), p->Font(), 
                                 p->Size(), p->Align(), p->Lines());
            break;
               
         case itemRectangle:
            DisplayRectangle(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent());
            break;
               
         case itemPresentTitle:
            DisplayPresentTitle(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemPresentSubtitle:
            DisplayPresentSubtitle(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemPresentTime:
            DisplayPresentTime(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemFollowingTitle:
            DisplayFollowingTitle(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemFollowingSubtitle:
            DisplayFollowingSubtitle(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemFollowingTime:
            DisplayFollowingTime(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemRecording:
            DisplayRecording(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemVolumeBackground:
            DisplayVolumeBackground(p->X(), p->Y(), p->Width(),p->Height(),p->Transparent(),p->Path());
            break;
               
         case itemVolumebar:
            DisplayVolumebar(p->X(), p->Y() ,p->Width(),p->Height(), p->Red(), p->Green(), p->Blue(),p->Transparent());
            break;
               
         case itemVolumeMuteSymbol:
            DisplayVolumeMuteSymbole(p->X(), p->Y(), p->Width(),p->Height(),p->Transparent(),p->Path(), p->Path2());
            break;
               
         case itemMessageBackground:
            DisplayMessageBackground(p->X(), p->Y(), p->Width(),p->Height(),p->Transparent(),p->Path());
            break;
               
         case itemMessage:
            DisplayMessage(p->X(), p->Y(), p->Width(), p->Height(), p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemReplayName:
            DisplayReplayName(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemReplayFolder:
            DisplayReplayFolder(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemReplayProcessbar:
            DisplayReplayProcessBar(p->X(), p->Y() ,p->Width(),p->Height(), 
                                    p->Red(), p->Green(), p->Blue(),
                                    p->Transparent(), p->BgX(), p->BgY(),
                                    p->BgWidth(),p->BgHeight(), 
                                    p->BgRed(), p->BgGreen(), p->BgBlue(),p->BgTransparent());
            break;
               
         case itemReplayCurrent:
            DisplayReplayCurrent(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(),p->Lines());
            break;
               
         case itemReplayTotal:
            DisplayReplayTotal(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemReplayRemaining:
            DisplayReplayRemaining(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemMenuTitle:
            DisplayMenuTitle(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;
               
         case itemMenuText:
            DisplayMenuText(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines());
            break;

         case itemMenuTextBackground:
            DisplayMenuTextBackground(p->X(), p->Y(), p->Width(),p->Height(), p->Transparent(), p->Path());
            break;

         case itemColumnSelected:
         case itemMenuSelected:
            pSelect = p;
            break;

         case itemMenu:
            DisplayMenuDefault(p, pSelect ? pSelect : p);
            break;

         case itemColumn:
            DisplayMenuColumn(p, pSelect ? pSelect : p);
            break;

         case itemMenuButtonRed:
            DisplayMenuButton(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines(), 0);
            break;

         case itemMenuButtonGreen:
            DisplayMenuButton(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines(), 1);
            break;

         case itemMenuButtonYellow:
            DisplayMenuButton(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines(), 2);
            break;

         case itemMenuButtonBlue:
            DisplayMenuButton(p->X(), p->Y(), p->Width(),p->Height(),p->Red(), p->Green(), p->Blue(), p->Transparent(), p->Font(), p->Size(), p->Align(), p->Lines(), 3);
            break;

         case itemMenuButtonBackgroundRed:
            DisplayMenuButtonBackground(p->X(), p->Y(), p->Width(),p->Height(), p->Transparent(),p->Path(), p->Path2(), 0);
            break;

         case itemMenuButtonBackgroundGreen:
            DisplayMenuButtonBackground(p->X(), p->Y(), p->Width(),p->Height(), p->Transparent(),p->Path(), p->Path2(), 1);
            break;

         case itemMenuButtonBackgroundYellow:
            DisplayMenuButtonBackground(p->X(), p->Y(), p->Width(),p->Height(), p->Transparent(),p->Path(), p->Path2(), 2);
            break;

         case itemMenuButtonBackgroundBlue:
            DisplayMenuButtonBackground(p->X(), p->Y(), p->Width(),p->Height(), p->Transparent(),p->Path(), p->Path2(), 3);
            break;

         case itemSpectrumAnalyzer:
            DisplaySpectrumAnalyzer(p);
            break;

         case itemCalibrationCursor:
            DisplayCalibrationCursor(p);
            break;

         default:
            break;
      }
   }
   
   DisplayShow();
}


void cGraphTFTDisplay::DisplayClear()
{
   _render->clear();
}

void cGraphTFTDisplay::DisplayShow()
{
   _render->refresh();

   if (comThread)
      comThread->refresh();

   if (GraphTFTSetup.DumpImage) 
   {
      int lastDumpDiff = _Dumped ? timeMs.Now() - _Dumped : 0;

      d(3, "DisplayShow: lastDumpDiff: %d  DumpRefresh: %d\n", lastDumpDiff, GraphTFTSetup.DumpRefresh * 100);

      if (lastDumpDiff > (GraphTFTSetup.DumpRefresh * 1000))
      {
         _Dumped =  timeMs.Now();
         _render->dumpImage2File("/graphTFT.png", GraphTFTSetup.DumpImageX, GraphTFTSetup.DumpImageY);
      }
   }
}

//***************************************************************************
// Display Routines for the Items
//***************************************************************************

void cGraphTFTDisplay::DisplayImage(int x, int y, int width, int height, int transparent, string image, int fit)
{
   fit = fit && width > 0 && height > 0;

   _render->image(image.c_str(), x, y, width, height, false, fit);
}

void cGraphTFTDisplay::DisplayImageCover(int x, int y, int width, int height, int transparent, string image)
{
   d(6,"DisplayImageCover: begin \n");

   if (GraphTFTSetup.CoverImage) 
      _render->image(GraphTFTSetup.CoverImage, x, y, width, height, true);
   else
      _render->image(image.c_str(), x, y, width, height, false);

   d(6,"DisplayImageCover: end \n");
}

void cGraphTFTDisplay::DisplayImageFile(int x, int y, int width, int height, int transparent, string file, string image)
{
   d(6,"DisplayImageFile\n");

   FILE *fp1;
   char oneline[100];
   char *c = 0;
   fp1 = fopen(file.c_str(), "r"); 

   if (fp1)
   {
      c = fgets(oneline, 100 ,fp1);

      if (c != 0)
      {
         oneline[strlen(oneline) -1 ] = '\0';
         d(6,"DisplayImageFile fgest: %s\n",oneline);
      }

      fclose(fp1);
   }

   d(6,"DisplayImageFile: fclosed\n");

   if (c != 0)
      _render->image(oneline, x, y, width, height, true);
   else
      _render->image(image.c_str(), x, y, width, height, false);
}

void cGraphTFTDisplay::DisplayMessageBackground(int x, int y, int width, int height, int transparent, string image)
{
   if (_messageIn <= 0)
   {
      _messageIn = 0;
      _message = "";
   }

   if (_message.length() > 0)
       DisplayImage(x, y, width, height, transparent, image);
}

void cGraphTFTDisplay::DisplayVolumeBackground(int x, int y, int width, int height, int transparent, string image)
{
   if (_volumeIn <= 0 || _mute)
      _volumeIn = 0;
   else
      DisplayImage(x, y, width, height, transparent, image);
}

void cGraphTFTDisplay::DisplayVolumeMuteSymbole(int x, int y, int width, int height, int transparent, string image, string image2)
{
   if (_mute && (GraphTFTSetup.ShowMutePermanent || _volumeIn > 0))
      DisplayImage(x, y, width, height, transparent, image);   // is Muted
   else if (image2 != "")
      DisplayImage(x, y, width, height, transparent, image2);
}

void cGraphTFTDisplay::DisplayDate(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   string date = *DayDateTime(time(0));
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, date ,false);
}

void cGraphTFTDisplay::DisplayDateOnly(int x, int y, int width, int height,int red, int green, int blue, 
                                       int transparent, string font, int size, int align, int lines)
{
   string date = *DayDateTime(time(0));
   date = date.substr(4,5);
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, date, false);
}

void cGraphTFTDisplay::DisplayTimeOnly(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   string date = *DayDateTime(time(0));
   date = date.substr(10,5);
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, date, false);
}

void cGraphTFTDisplay::DisplayDateTimeFormat(cGraphTFTThemeItem* p)
{
   char date[500];
   time_t t;
   tm* tmp;

   t = time(0);
   tmp = localtime(&t);

   if (!tmp) 
   {
      d(0, "Error: Can't get localtime!");
      return ;
   }

   if (strftime(date, sizeof(date), p->Format().c_str(), tmp) == 0)
   {
      d(0, "Error: Can't convert time, maybe invalid format string '%s'!", p->Format().c_str());
      return ;
   }
   
   DisplayText(p->X(), p->Y(), p->Width(), p->Height(),
               p->Red(), p->Green(), p->Blue(), p->Transparent(), 
               p->Font(), p->Size(), p->Align(), p->Lines(), 
               date, false);
}

void cGraphTFTDisplay::DisplayText(int x, int y, int width, int height,int red, int green, int blue, 
                                   int transparent, string font, int size, int align, 
                                   int lines, string text, bool marquee)
{
   string title_text;

   if (!width)   width = 720 - x;
   if (!height)  height = 576 - y;

   if (!marquee || GraphTFTSetup.ViewType < 2)
   {
      // Normal Modus und gekuerzter Modus

      _render->text(text.c_str(), font.c_str(), size, align, x, y, 
                    red, green, blue, width, height, lines,
                    marquee && GraphTFTSetup.ViewType == 1 ? true : false);

      return ;
   }

   if (GraphTFTSetup.ViewType == 2)
   {
      // Marquee Modus

      if ((int)text.length() <= (int)GraphTFTSetup.ViewLength) 
      {
         _render->text(text.c_str(), font.c_str(), size, align, x, y, red, green, blue, width, height, lines);
      } 
      else 
      {
         if (_marquee_active == false) 
         {
            _marquee_active = true;
            _marquee_left   = false;
            _marquee_idx    = 0;
            _marquee_count  = 0;
         } 
         else 
         {
            if (GraphTFTSetup.ScrollCount == 0)
            {
               if ((int)(_marquee_idx + GraphTFTSetup.ViewLength) < (int)text.length() && _marquee_left==false)
               {
                  _marquee_idx ++;
               } 
               else 
               {
                  _marquee_left = true;

                  if (_marquee_idx > 0) 
                     _marquee_idx --;
                  else 
                     _marquee_left = false;
               }
            }
            else if (_marquee_count < GraphTFTSetup.ScrollCount) 
            {
               if ((int)(_marquee_idx + GraphTFTSetup.ViewLength) < (int)text.length() && _marquee_left==false)
               {
                  _marquee_idx ++;
               } 
               else 
               {
                  _marquee_left = true;
                        
                  if (_marquee_idx > 0) 
                  {
                     _marquee_idx--;

                     if (_marquee_idx == 0)
                     {
                        _marquee_left = false;
                        _marquee_count++;
                     }
                  }
               }
            } 
            else 
            {
               _marquee_active = false;
            }
         }

         title_text = text.substr(_marquee_idx, GraphTFTSetup.ViewLength);
         _render->text(title_text.c_str(), font.c_str(), size, align, x, y, red, green, blue, width, height, lines);

         if (_marquee_idx == 1) 
            usleep(GraphTFTSetup.ViewLength * 40000);
      }
   }

   else if (GraphTFTSetup.ViewType == 3)
   {
      // Ticker Modus

      if ((int)text.length() <= (int)GraphTFTSetup.ViewLength)
      {
         _render->text(text.c_str(), font.c_str(), size, align, x, y, red, green, blue, width, height, lines);
      } 
      else 
      {
         if (_marquee_active == false) 
         {
            _marquee_active = true;
            _marquee_idx    = 0;
            _marquee_count  = 0;
            _marquee_strip  = 0;
         } 
         else 
         {
            if (GraphTFTSetup.ScrollCount == 0)
            {
               // endlos tickern

               if ((int)(_marquee_idx + GraphTFTSetup.ViewLength) < (int)text.length())
               {
                  _marquee_idx ++;
               } 
               else 
               {
                  if (_marquee_idx < (int)text.length()) 
                  {
                     _marquee_idx ++;
                     _marquee_strip ++;
                  } 
                  else 
                  {
                     _marquee_active = false;
                  }
               }
            } 

            else if (_marquee_count < GraphTFTSetup.ScrollCount) 
            {
               // Anzahl ScrollCount tickern

               if ((int)(_marquee_idx + GraphTFTSetup.ViewLength) < (int)text.length())
               {
                  _marquee_idx ++;
               } 
               else 
               {
                  if (_marquee_idx < (int)text.length())
                  {
                     _marquee_idx ++;
                     _marquee_strip ++;
                  } 
                  else 
                  {
                     _marquee_idx    = 0;
                     _marquee_count++;
                     _marquee_strip  = 0;
                  }
               }
            } 
            else 
            {
               _marquee_active = false;
            }
         }

         title_text = text.substr(_marquee_idx, (GraphTFTSetup.ViewLength-_marquee_strip));
         _render->text(title_text.c_str(), font.c_str(), size, align, x, y, red, green, blue, width, height, lines);

         if (_marquee_idx == 1) 
            usleep(GraphTFTSetup.ViewLength * 40000);
      }
   }
}

void cGraphTFTDisplay::DisplayChannelName(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   string text = _channelInfo != "" ? _channelInfo : _channelText;
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, skipspace(text.c_str()), false);
}

void cGraphTFTDisplay::DisplayChannelNumber(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   string text = _channelInfo != "" ? _channelInfo : _channelText;
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, text.substr(0, text.find(' ')), false);
}

void cGraphTFTDisplay::DisplayRectangle(int x, int y, int width, int height,int red, int green, int blue, int transparent)
{
   _render->rectangle(x, y, width, height, red, green, blue, transparent);
}

void cGraphTFTDisplay::DisplayChannelNameOnly(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   string text = _channelInfo != "" ? _channelInfo : _channelText;
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, text.substr(text.find(' ') + 1), false);
}

void cGraphTFTDisplay::DisplayLogo(int x, int y, int width, int height, int transparent, string image, string type, int animated, int delay, int count)
{
   string path;

   if (isModeNormal(_mode) && type != "") 
   {
      string text = _channelInfo != "" ? _channelInfo : _channelText;
      path = image.substr(0,image.rfind('/') + 1) + text.substr(text.find(' ') + 1) + "." + type;
      std::ostringstream fullpath; fullpath << GraphTFTSetup.PluginConfPath << "/graphTFT/themes/" << theTheme->getDir() << "/" << path;

      if (!file_exists(fullpath.str())) 
         path = image;
   } 
   else
      path = image;

   if (animated) 
   {
      if (!_animation) 
      {
         _animation = true;
         _displayFrame = 0;
         _frameIn = delay;
      } 
      else if (_frameIn < 0) 
      {
         _displayFrame = (_displayFrame + 1) % count;
         _frameIn = delay;
      }

      int pos;

      if ((pos = path.rfind('.')) != -1) 
      {
         char *temp;
         asprintf(&temp, "%s%d%s", path.substr(0, pos).c_str(), _displayFrame, path.substr(pos).c_str());
         path = temp;
         free(temp);
      }
   } 
   else
      _animation = false;

   d(2, "looking for image '%s'", path.c_str());
   
   DisplayImage(x, y, width, height, transparent, path, true);
}

void cGraphTFTDisplay::DisplayPresentTitle(int x, int y, int width, int height,int red, int green, int blue, 
                                           int transparent, string font, int size, int align, int lines)
{
   if (_channelInfo == "" || _channelInfo == _channelText)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _presentTitle, true);
}

void cGraphTFTDisplay::DisplayPresentTime(int x, int y, int width, int height,int red, int green, int blue, 
                                          int transparent, string font, int size, int align, int lines)
{
   string presentTime = *DayDateTime(_presentTime);
   presentTime = presentTime.substr(10);

   if (_channelInfo == "" || _channelInfo == _channelText)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, presentTime, false);
}

void cGraphTFTDisplay::DisplayPresentSubtitle(int x, int y, int width, int height,int red, int green, int blue, 
                                              int transparent, string font, int size, int align, int lines)
{
   if (_channelInfo == "" || _channelInfo == _channelText)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _presentSubtitle, false);
}

void cGraphTFTDisplay::DisplayFollowingTitle(int x, int y, int width, int height,int red, int green, int blue, 
                                             int transparent, string font, int size, int align, int lines)
{
   if (_channelInfo == "" || _channelInfo == _channelText) 
   {
      if ((int)_followingTitle.length() <= (int)GraphTFTSetup.ViewLength) 
         DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _followingTitle, true);
      else 
         DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _followingTitle, false);
   }
}

void cGraphTFTDisplay::DisplayFollowingSubtitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   if (_channelInfo == "" || _channelInfo == _channelText)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _followingSubtitle, false);
}

void cGraphTFTDisplay::DisplayFollowingTime(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   string followingTime = *DayDateTime(_followingTime);
   followingTime = followingTime.substr(10);

   if (_channelInfo == "" || _channelInfo == _channelText)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, followingTime, false);
}

void cGraphTFTDisplay::DisplayTimebar(cGraphTFTThemeItem* item)
{
   float percent;
   int xDone;
   int red = item->Red();
   int green = item->Green();
   int blue = item->Blue();

   if ((_channelInfo == "" || _channelInfo == _channelText) && _followingTime && _followingTime != _presentTime) 
   {
      percent = ((float)(time(0) - _presentTime)) / ((float)(_followingTime - _presentTime));

      if (percent < 0)   percent = 0;
      if (percent > 1.0) percent = 1.0;

      xDone = (int)(percent*(float)(item->Width() > item->Height() ? item->Width() : item->Height()));

      // the background 

      _render->rectangle(item->BgX(), item->BgY(), 
                         item->BgWidth(), item->BgHeight(), 
                         item->BgRed(), item->BgGreen(), item->BgBlue(), 
                         item->BgTransparent());
      
      if (item->Path() != "")
      {
         // timebar with image
         
         _render->image(item->Path().c_str(), 
                        item->X(), item->Y(),
                        item->Width(), item->Height(),
                        false, true);

         _render->rectangle(item->X()+xDone, item->Y(), 
                            item->Width()-xDone, item->Height(),
                            item->BgRed(), item->BgGreen(), item->BgBlue(), 
                            item->BgTransparent());
      }

      else 
      {
         if (item->Switch())
         {
            // colorchanging timebar

            red = green = 255;
            blue = 0;

            if (percent < 0.5f) 
               red = (int)(255.0f * percent * 2.0f);
            else 
               green = (int)(255.0f * (1-percent) * 2.0f);
         }

         // foreground

         if (item->Width() > item->Height())
            _render->rectangle(item->X(), item->Y(), 
                               xDone, item->Height(), 
                               red, green, blue,
                               item->Transparent());
         else
            _render->rectangle(item->X(), item->Y(), 
                               item->Width(), xDone, 
                               red, green, blue,
                               item->Transparent());
      }
   }
}

void cGraphTFTDisplay::DisplayMessage(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   if (_messageIn < 0)
   {
      _message = "";
      _messageIn = 0;
   }

   if (_message.length() > 0)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _message, false);
}

void cGraphTFTDisplay::DisplayRecording(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   if (_recordings.size() > 0)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _recordings[_showRecording].name.c_str(), false);
}

void cGraphTFTDisplay::DisplaySymbols(int x, int y, int width, int height, int transparent, string path, string path2, int symbol)
{
   const char *image = 0;
   bool play, forward;
   int speed;
   cChannel *chan;
   
   if (_replay.control && _replay.control->GetReplayMode(play, forward, speed))
   {
      switch (symbol) 
      {
         case itemSymRepPlay:
            image = (play && speed == -1) ? path.c_str() : path2.c_str();
            break;
         case itemSymRepPause:
            image = (!play && speed == -1) ? path.c_str() : path2.c_str();
            break;
         case itemSymRepFF:
            image = (play && forward && speed != -1) ? path.c_str() : path2.c_str();
            break;
         case itemSymRepFB:
            image = (play && !forward && speed != -1) ? path.c_str() : path2.c_str();
            break;
         case itemSymRepSF:
            image = (!play && forward && speed != -1) ? path.c_str() : path2.c_str();
            break;
         case itemSymRepSB:
            image = (!play && !forward && speed != -1) ? path.c_str() : path2.c_str();
            break;
      }
   } 
   else if (isModeNormal(_mode) && (chan = Channels.GetByNumber(_channel))) 
   {
      switch (symbol) 
      {
         case itemSym2ch:
            image = chan->Apid(1) ? path.c_str() : path2.c_str();
            break;
         case itemSymDD:
            image = chan->Dpid(0) ? path.c_str() : path2.c_str();
            break;
         case itemSymVTX:
            image = chan->Tpid() ? path.c_str() : path2.c_str();
            break;
         case itemSymCrypt:
            image = chan->Ca() ? path.c_str() : path2.c_str();
            break;
      }
   }

   if (image && *image)
      _render->image(image, x, y, width, height, false);
}

void cGraphTFTDisplay::DisplayRecordingSymbol(int x, int y, int width, int height, int transparent, string path, string path2)
{
   if (_recordings.size() > 0)
      _render->image(path.c_str(), x, y, width, height, false);
   else if (path2.length() > 0)
      _render->image(path2.c_str(), x, y, width, height, false);
}

void cGraphTFTDisplay::DisplayMenuTitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _menu.title.c_str(), false);
}

void cGraphTFTDisplay::DisplayMenuText(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   if (_menu.text.length() != 0) 
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _menu.text.c_str(), false);
}

void cGraphTFTDisplay::DisplayMenuTextBackground(int x, int y, int width, int height, int transparent, string image)
{
   if (_menu.text.length() != 0) 
      DisplayImage(x, y, width, height, transparent, image);
}

void cGraphTFTDisplay::DisplayMenuDefault(cGraphTFTThemeItem* item, cGraphTFTThemeItem* selectedItem)
{
   int tabChars;
   int afterSelect = 0;
   int lineHeight = item->Size() * 5/3;
   int lineHeightSelect = selectedItem->Size() * 5/3;
   int count = ((item->Height()-lineHeightSelect) / lineHeight) + 1;
   int step = _menu.currentPos - _menu.currentPosLast;       // step since last refresh

   _menu.lineHeight = lineHeight;

   if (_menu.topPos < 0)
      _menu.topPos = max(0, _menu.currentPos - count/2);     // initial
   else if (_menu.currentPos == _menu.topPos-1)
      _menu.topPos = _menu.currentPos;                       // up
   else if (_menu.currentPos == _menu.topPos+count)
      _menu.topPos++;                                        // down
   else if (_menu.currentPos < _menu.topPos 
            || _menu.currentPos > _menu.topPos+count)
      _menu.topPos += step;                                  // page up / page down

   if (_menu.topPos > (int)_menu.items.size()-count)
      _menu.topPos = _menu.items.size()-count;

   if (_menu.topPos < 0)
      _menu.topPos = 0;

   _menu.currentPosLast = _menu.currentPos;

   // calculate column width

   tabChars = charInTabs[0] + charInTabs[1] + charInTabs[2] + charInTabs[3] + charInTabs[4] + charInTabs[5];

   if (tabChars > 0)
   {
      int saveTabsProz[7] = {0,0,0,0,0,0,0};
      int total = 100;
      
      for (int i = 0; charInTabs[i] != 0; i++)
      {
         if (charInTabs[i] < (tabChars/6))
            saveTabsProz[i] = (int)(((float)total / (float)item->Width()) * (float)item->Size() * ((float)charInTabs[i]+1));
         else
            saveTabsProz[i] = (total * charInTabs[i]) / tabChars;
         
         total -= saveTabsProz[i];      
      }
      
      SetTabs(saveTabsProz[0], saveTabsProz[1], saveTabsProz[2], saveTabsProz[3], saveTabsProz[4]);
      
      d(5, "Graphtft: tabs set to - %d:%d:%d:%d:%d", 
        saveTabsProz[0],saveTabsProz[1],saveTabsProz[2],saveTabsProz[3],saveTabsProz[4]);
   }
   
   else
      SetTabs(0,0,0,0,0);
   
   // loop over visible rows ...

   for (int i = _menu.topPos; i < min((int)_menu.items.size(), _menu.topPos + count); ++i) 
   {
      cGraphTFTThemeItem* p = i == _menu.currentPos ? selectedItem : item;
      int y = p->Y() + ((i - _menu.topPos) * lineHeight);

      if (i == _menu.currentPos)
      {
         afterSelect = lineHeightSelect - lineHeight;

         // draw the selected backround

         if (selectedItem->Path() != "")
            _render->image(selectedItem->Path().c_str(), item->X(), item->Y() + (i - _menu.topPos) * lineHeight, item->Width(), item->Height(), false);

         // draw the columns

         for (int t = 0; t < MaxTabs; ++t) 
         {
            if (_menu.items[i].tabs[t] != "") 
               _render->text(_menu.items[i].tabs[t].c_str(), 
                             selectedItem->Font().c_str(), selectedItem->Size(), selectedItem->Align(),
                             item->X() + int(item->Width() * Tab(t) / 100), 
                             item->Y() + ((i - _menu.topPos) * lineHeight), 
                             selectedItem->Red(), selectedItem->Green(), selectedItem->Blue(), 
                             int(item->Width() * (Tab(t+1)-Tab(t)) / 100), 
                             lineHeightSelect, 
                             1);

            if (!Tab(t + 1) || Tab(t+1) == 100)
               break;
         }

         if (selectedItem->Path2() != "")
            _render->image(selectedItem->Path2().c_str(), 
                           item->X(), item->Y() + (i - _menu.topPos) * lineHeight, 
                           item->Width(), item->Height(), false);

         if (p->StaticPicture())
         {
            // for the selected item we optionaly draw a additional picture (ImageMap)
            // as image name use the text after the number in the last column ...

            string path = currentSection->getPathFromImageMap(_menu.items[i].tabs[_menu.items[i].tabCount-1].c_str());

            if (path != "")
            {
               if (p->StaticX() && p->StaticY())
                  DisplayImage(p->StaticX(), p->StaticY(), 128, 128, 255, path);
               else if (p->StaticX())
                  DisplayImage(p->StaticX(), y, 128, 128, 255, path);
               else
                  DisplayImage(720-128, y, 128, 128, 255, path);   // historical for old themes :(
            }
         }
      }

      else
      {   
         if (item->Path() != "")
            _render->image(item->Path().c_str(), item->X(), y + afterSelect, item->Width(), item->Height(), false);

         for (int t = 0; t < MaxTabs; ++t) 
         {
            if (_menu.items[i].tabs[t] != "") 
               _render->text(_menu.items[i].tabs[t].c_str(),                      // test, 
                             item->Font().c_str(), item->Size(), item->Align(),   // font, font-size, align
                             item->X() + int(item->Width() * Tab(t) / 100),       // x-pos
                             y + afterSelect,                                     // y-pos
                             item->Red(), item->Green(), item->Blue(),            // RGB
                             int(item->Width() * (Tab(t+1)-Tab(t)) / 100),        // width
                             lineHeight,                                          // height
                             1);                                                  // line-count
               
            if (!Tab(t + 1) || Tab(t+1) == 100)
               break;
         }
      }
   }
}

//***************************************************************************
// Display Menu Column
//***************************************************************************

void cGraphTFTDisplay::DisplayMenuColumn(cGraphTFTThemeItem* item, 
                                         cGraphTFTThemeItem* selectedItem)
{
   int afterSelect = 0;
   int lineHeight = item->Size() * 5/3;
   int lineHeightSelect = selectedItem->Size() * 5/3;
   int count = ((item->Height()-lineHeightSelect) / lineHeight) + 1;
   int step = _menu.currentPos - _menu.currentPosLast;       // step since last refresh

   _menu.lineHeight = lineHeight;

   if (_menu.topPos < 0)
      _menu.topPos = max(0, _menu.currentPos - count/2);     // initial
   else if (_menu.currentPos == _menu.topPos-1)
      _menu.topPos = _menu.currentPos;                       // up
   else if (_menu.currentPos == _menu.topPos+count)
      _menu.topPos++;                                        // down
   else if (_menu.currentPos < _menu.topPos
            || _menu.currentPos > _menu.topPos+count)
      _menu.topPos += step;                                  // page up / page down
   
   if (_menu.topPos > (int)_menu.items.size()-count)
      _menu.topPos = _menu.items.size()-count;

   if (_menu.topPos < 0)
      _menu.topPos = 0;
   
   _menu.currentPosLast = _menu.currentPos;

   // paint the visible rows

   for (int i = _menu.topPos; i < min((int)_menu.items.size(), _menu.topPos + count); ++i) 
   {
      cGraphTFTThemeItem* p = i == _menu.currentPos ? selectedItem : item;
      int y = p->Y() + ((i - _menu.topPos) * lineHeight) + afterSelect;

      if (i == _menu.currentPos)
      {
         afterSelect = lineHeightSelect - lineHeight;
         
         if (p->ImageMap())
         {
            // for the selected item we optionaly draw a additional picture (ImageMap)
            // as image name use the text after the number in the last column ...

            string path = currentSection->getPathFromImageMap(_menu.items[i].tabs[p->Number()-1].c_str());

            if (path != "")
            {
               if (p->StaticX() && p->StaticY())
                  DisplayImage(p->StaticX(), p->StaticY(), 128, 128, 255, path);
               else if (p->StaticX())
                  DisplayImage(p->StaticX(), y, 128, 128, 255, path);
            }
         }

         if (p->StaticPicture())
         {
            // Image with static position for this column

            if (_menu.items[i].tabs[p->Number()-1].c_str()[0])
            {
               string path = "columnimages/" + _menu.items[i].tabs[p->Number()-1] + p->Format();
            
               d(2, "looking for image '%s'", path.c_str());

               if (p->StaticX() && p->StaticY())
                  _render->image(path.c_str(), p->StaticX(), p->StaticY(), 0, 0, false, false);
               else if (p->StaticX())
                  _render->image(path.c_str(), p->StaticX(), y, 0, 0, false, false);
            }
         }

         if (p->StaticText())
         {
            // Text with static position for this column

            if (_menu.items[i].tabs[p->Number()-1].c_str()[0])
            {
               _render->text(_menu.items[i].tabs[p->Number()-1].c_str(),    // text
                             p->Font().c_str(), p->Size(), p->Align(),      // font, font-size, align
                             p->StaticX(),                                  // x-pos
                             p->StaticY(),                                  // y-pos
                             p->Red(), p->Green(), p->Blue(),               // RGB
                             p->StaticWidth(),                              // width
                             lineHeight, 1);                                // height, line-count
            }
         }
      }

      if (p->Width() == -1)
         continue;
      
      cGraphTFTThemeItem* partingLine;

      if (_menu.items[i].type == itPartingLine && (partingLine = currentSection->getItem(itemPartingLine)))
      {
         DisplayPartingLine(partingLine, _menu.items[i].tabs[0], y, lineHeight);
         
         continue;
      }

      // draw image
      
      if (p->Path() != "")
         _render->image(p->Path().c_str(), 
                        p->X(), y, 
                        p->Width(), lineHeight, false);
      
      if (p->Type() == "progress")
      {
         int current = 0;
         int n = 0;
         int barHeight = p->BarHeight();
         
         if (p->BarHeightUnit() == iuPercent)      // BarHeight in % of row
            barHeight = (int)(((double)lineHeight/100) * (double)p->BarHeight());

         // calc progress

         while (_menu.items[i].tabs[p->Number()-1][n])
         {
            if (_menu.items[i].tabs[p->Number()-1][n] == '|')
               current++;
            n++;
         }

         // draw column's progress-bar
         
         d(5, "progress [%d] of (%d%%) at position y = %d [%s]", 
           i, current, y, _menu.items[i].tabs[p->Number()-1].c_str());

         DisplayProgressBar(p, current, 8, y + (lineHeight-barHeight)/2, barHeight);
      }

      else if (p->Type() == "image")
      {
         // draw column's image

         if (_menu.items[i].tabs[p->Number()-1].length() > 0)
         {
            // string path = "columnimages/" + _menu.items[i].tabs[p->Number()-1] + p->Format();

            int len = _menu.items[i].tabs[p->Number()-1].length();
            char* file = new char[len+1];
            char* path = new char[len+strlen("columnimages/.")+strlen(p->Format().c_str())+1];
            int pos = 0;
            
            while (_menu.items[i].tabs[p->Number()-1].c_str()[pos])
            {
               if (!strchr("/'", _menu.items[i].tabs[p->Number()-1].c_str()[pos]))
                  file[pos] = _menu.items[i].tabs[p->Number()-1].c_str()[pos];

               pos++;
            }
            
            file[pos] = 0;
            sprintf(path, "columnimages/%s.%s", file, p->Format().c_str());

            d(2, "looking for image '%s'", path);

            if (p->BgRed() || p->BgGreen() || p->BgBlue())
               _render->rectangle(p->X(), y, p->Width(), lineHeight, 
                                  p->BgRed(), p->BgGreen(), p->BgBlue(), 
                                  p->BgTransparent());
         
            if (_menu.items[i].tabs[p->Number()-1] != "")
            {
               _render->image(path, 
                              p->X(), y, 
                              p->Width(), lineHeight, false, true);
            }

            delete[] path;
            delete[] file;
         }
      }

      else if (_menu.items[i].tabs[p->Number()-1] != "")
      {
         // draw column's text

         _render->text(_menu.items[i].tabs[p->Number()-1].c_str(),    // text
                       p->Font().c_str(), p->Size(), p->Align(),      // font, font-size, align
                       p->X(),                                        // x-pos
                       y,                                             // y-pos
                       p->Red(), p->Green(), p->Blue(),               // RGB
                       p->Width() ? p->Width() : 720 - p->X(),        // width
                       lineHeight, 1, true);                          // height, line-count
      }
   }
}

//***************************************************************************
// Display Partingline
//***************************************************************************

void cGraphTFTDisplay::DisplayPartingLine(cGraphTFTThemeItem* p, string text, int y, int height)
{
   if (!p)
      return;

   if (p->Path() != "" && text != "")
      _render->image(p->Path().c_str(), 
                     p->X(), y, 
                     p->Width(), height, false);

   if (p->Path2() != "" && text == "")
      _render->image(p->Path2().c_str(), 
                     p->X(), y, 
                     p->Width(), height, false);
   
   if (text != "")
      _render->text(text.c_str(),    // text
                    p->Font().c_str(), p->Size(), p->Align(),      // font, font-size, align
                    p->X(),                                        // x-pos
                    y,                                             // y-pos
                    p->Red(), p->Green(), p->Blue(),               // RGB
                    p->Width(),                                    // width
                    height, 1);                                    // height, line-count
   
}

void cGraphTFTDisplay::DisplayMenuButton(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines, int buttonNumber)
{
   if (_menu.buttons[buttonNumber].length() > 0)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _menu.buttons[buttonNumber].c_str(), false);
}

void cGraphTFTDisplay::DisplayMenuButtonBackground(int x, int y, int width, int height, int transparent, string image, string imageOff, int buttonNumber)
{
   if (_menu.buttons[buttonNumber].length() > 0)
      DisplayImage(x, y, width, height, transparent, image);
   else
      DisplayImage(x, y, width, height, transparent, imageOff);
}

void cGraphTFTDisplay::DisplayVolumebar(int x, int y ,int width, int height, int red, int green, int blue, int transparent)
{
   int xVolume = (_volume * (width > height ? width : height)) / 255;

   if (_volumeIn > 0 && !_mute) 
   {
      _render->rectangle(x, y, width, height, 0, 0, 0, transparent);

      if (width > height) 
         _render->rectangle(x, y, xVolume, height, red, green, blue, transparent);
      else
         _render->rectangle(x, y, width, xVolume, red, green, blue, transparent);
   }
}

void cGraphTFTDisplay::DisplayReplayName(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   if (_replay.name.length() > 0)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _replay.name, false);
}

void cGraphTFTDisplay::DisplayReplayFolder(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   if (_replay.folder.length() > 0)
      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, _replay.folder, false);
}

void cGraphTFTDisplay::DisplayReplayProcessBar(int x, int y ,int width, int height, int red, int green, int blue, int transparent, int bg_x, int bg_y, int bg_width, int bg_height, int bg_red, int bg_green, int bg_blue, int bg_transparent)
{
   int current, total;

   if (_replay.control && _replay.control->GetIndex(current, total)) 
   {
      if (total == 0 && current == 0)
         current = total = 1;
      else if (total == 0)
         total = current;

      int xDone= (int)(((float)current / (float)total)*(float)(width > height ? width : height));

      if (bg_x == 0)      bg_x = x;
      if (bg_y == 0)      bg_y = y;
      if (bg_width == 0)  bg_width = width;
      if (bg_height == 0) bg_height = height;

      _render->rectangle(bg_x, bg_y, bg_width, bg_height, bg_red, bg_green, bg_blue, bg_transparent);

      if (width > height)
         _render->rectangle(x, y, xDone, height, red, green, blue, transparent);
      else
         _render->rectangle(x, y, width, xDone, red, green, blue, transparent);
   }
}

//***************************************************************************
// Display Progress Bar
//***************************************************************************

void cGraphTFTDisplay::DisplayProgressBar(cGraphTFTThemeItem* p, float current,
                                          float total, int y, int height)
{
   int xDone;

   height = height == -1 ? p->Height() : height;
   y = y == -1 ? p->Y() : y;
   xDone = (int)((current/total) * (float)p->Width());
   
   // background

   if (p->Path2() != "")
      _render->image(p->Path2().c_str(), 
                     p->X(), y,
                     p->Width(), height, 
                     false, true);
   else
      _render->rectangle(p->X(), y, p->Width(), height,
                         p->BgRed(), p->BgGreen(), p->BgBlue(), 
                         p->BgTransparent());
   
   // foreground

   if (p->Path2() != "")
      _render->rectangle(p->X()+xDone, y, 
                         p->Width()-xDone, height,
                         p->BgRed(), p->BgGreen(), p->BgBlue(),
                         p->BgTransparent());
   else
      _render->rectangle(p->X(), y, xDone, height,
                         p->Red(), p->Green(), p->Blue(), 
                         p->Transparent());
}

void cGraphTFTDisplay::DisplayReplayCurrent(int x, int y, int width, int height, 
                                            int red, int green, int blue, int transparent, 
                                            string font, int size, int align, int lines)
{
   int current, total;

   if (_replay.control && _replay.control->GetIndex(current, total)) 
   {
      if (total == 0 && current == 0)
         current = total = 1;

      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, *IndexToHMSF(current), false);
   }
}

void cGraphTFTDisplay::DisplayReplayTotal(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   int current, total;

   if (_replay.control && _replay.control->GetIndex(current, total)) 
   {
      if (total == 0 && current == 0)
         current = total = 1;
      else if (total == 0)
         total = current;

      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, *IndexToHMSF(total), false);
   }
}

void cGraphTFTDisplay::DisplayReplayRemaining(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines)
{
   int current, total;

   if (_replay.control && _replay.control->GetIndex(current, total))
   {
      if (total == 0 && current == 0)
         current = total = 1;
      else if (total == 0)
         total = current;

      DisplayText(x, y, width, height, red, green, blue, transparent, font, size, align, lines, (string)"-" + *IndexToHMSF(total - current), false);
   }
}

//***************************************************************************
// Display Calibration Cursor
//***************************************************************************

void cGraphTFTDisplay::DisplayCalibrationCursor(cGraphTFTThemeItem* p)
{
   if (p->Path() != "")
   {
      _render->image(p->Path().c_str(), 
                     calibrationPos.x, calibrationPos.y,
                     p->Width(), p->Height(),
                     false, p->Width() || p->Height());     
   }
   else
   {
      _render->rectangle(calibrationPos.x, calibrationPos.y, 
                         p->Width() ? p->Width() : 10, 
                         p->Height() ? p->Height() : 10,
                         p->Red(), p->Green(), p->Blue(), 
                         p->Transparent());
   }
}

//***************************************************************************
// Display Spectrum Analyzer
//***************************************************************************

void cGraphTFTDisplay::DisplaySpectrumAnalyzer(cGraphTFTThemeItem* item)
{

	if (GraphTFTSetup.enableSpectrumAnalyzer)
	{
		if (cPluginManager::CallFirstService(SPAN_GET_BAR_HEIGHTS_ID, 0))
		{
			Span_GetBarHeights_v1_0 GetBarHeights;
			
			int bandsSA = 20;
			int falloffSA = 8;
			
			unsigned int* barHeights = new unsigned int[bandsSA];
			unsigned int* barHeightsLeftChannel = new unsigned int[bandsSA];
			unsigned int* barHeightsRightChannel = new unsigned int[bandsSA];
			unsigned int* barPeaksBothChannels = new unsigned int[bandsSA];
			unsigned int* barPeaksLeftChannel = new unsigned int[bandsSA];
			unsigned int* barPeaksRightChannel = new unsigned int[bandsSA];
			unsigned int volumeLeftChannel;
			unsigned int volumeRightChannel;
			unsigned int volumeBothChannels;
							
			GetBarHeights.bands                  = bandsSA;
			GetBarHeights.barHeights				 = barHeights;
			GetBarHeights.barHeightsLeftChannel  = barHeightsLeftChannel;
			GetBarHeights.barHeightsRightChannel = barHeightsRightChannel;
			GetBarHeights.volumeLeftChannel		 = &volumeLeftChannel;
			GetBarHeights.volumeRightChannel		 = &volumeRightChannel;
			GetBarHeights.volumeBothChannels		 = &volumeBothChannels;

			GetBarHeights.name						 = "graphtft";
			GetBarHeights.falloff					 = falloffSA;
			GetBarHeights.barPeaksBothChannels	 = barPeaksBothChannels;
			GetBarHeights.barPeaksLeftChannel    = barPeaksLeftChannel;
			GetBarHeights.barPeaksRightChannel	 = barPeaksRightChannel;
			
			if (cPluginManager::CallFirstService(SPAN_GET_BAR_HEIGHTS_ID, &GetBarHeights))
			{

            // 				int saStartX = item->X();
            // 				// int saEndX = item->X() + item->Width();     // saStartX + barWidth*bandsSA*2 + bandsSA/4 - 1;
            // 				int saStartY = item->Y();
            // 				int saEndY = item->Y() + item->Height();    // item->Y() + bitmap->Height()/2 - 3;
				
            // 				// LastTimeSA.Set(100);
				
            // 				// if (GraphTFTSetup.SAShowVolume)
            // 				{
					
            // 					saStartX = FRAME_SPACE_X  + bitmap->Width()/2 - (barWidth*bandsSA*2 + bandsSA/4)/2 - 2;
            // 					saEndX = saStartX + barWidth*bandsSA*2 + bandsSA/4 - 1;
					
            // 					// left volume

            // 					bitmap->DrawRectangle(FRAME_SPACE_X,
            //                                      saStartY,
            //                                      saStartX-1,
            //                                      saEndY + 1,
            //                                      GLCD::clrWhite, true);
					
            // 					for ( i=0; (i<logo->Width()/2-2) && (i<3*(volumeLeftChannel*saStartX)/100); i++)
            // 					{
            // 						bitmap->DrawRectangle(saStartX - i - 2,
            //                                         saStartY + saEndY/2 - i,
            //                                         saStartX - i - 4,
            //                                         saStartY + saEndY/2 + i,
            //                                         GLCD::clrBlack, true);
            // 					}
					
            // 					// right volume

            // 					bitmap->DrawRectangle(saEndX + 1,
            //                                      saStartY,
            //                                      bitmap->Width() - 1,
            //                                      saEndY + 1,
            //                                      GLCD::clrWhite, true);
					
            // 					for ( i=0; (i<logo->Width()/2-2) && (i<3*(volumeRightChannel*saStartX)/100); i++)
            // 					{
            // 						bitmap->DrawRectangle(saEndX + 2 + i,
            //                                         saStartY + saEndY/2 - i,
            //                                         saEndX + i + 4,
            //                                         saStartY + saEndY/2 + i,
            //                                         GLCD::clrBlack, true);
            // 					}
            // 				}

				// black background

            // 				bitmap->DrawRectangle(saStartX,
            //                                   saStartY,
            //                                   saEndX,
            //                                   saEndY + 1,
            //                                   GLCD::clrBlack, true);

				int i;
				int barWidth = item->Width() / (2 * bandsSA);
            int width = barWidth * 2 * bandsSA;

            d(4, "width = %d; barWidth = %d; ", width, barWidth);

            if (item->Path() != "")
               _render->image(item->Path().c_str(), 
                              item->X(), item->Y(),
                              width, item->Height(), false, true);
            else
               _render->rectangle(item->X(), item->Y(),
                                  width, item->Height(),
                                  item->Red(), item->Green(), item->Blue(),
                                  item->Transparent());
            
				for (i = 0; i < bandsSA; i++)
				{
               _render->rectangle(item->X() + 2 * barWidth * i,
                                  item->Y(),
                                  barWidth,
                                  item->Height() - (barHeightsLeftChannel[i] * item->Height() / 100),
                                  0, 0, 0,
                                  item->Transparent());

               _render->rectangle(item->X() + 2 * barWidth * i + barWidth,
                                  item->Y(),
                                  barWidth,
                                  item->Height() - (barHeightsRightChannel[i] * item->Height() / 100),
                                  0, 0, 0,
                                  item->Transparent());
						
               // the peak

//                height = barPeaksBothChannels[i] * item->Height() / 100;

//                if (height > 0)
//                {
//                   _render->rectangle(item->X() + barWidth*2*i+ barWidth + 1,
//                                      item->Y(),
//                                      item->X() + barWidth*2*i + barWidth+ barWidth + 1,
//                                      height,
//                                      item->Red(), item->Green(), item->Blue(),
//                                      item->Transparent());
//                }
            }
         }
         
         delete[] barHeights;
         delete[] barHeightsLeftChannel;
         delete[] barHeightsRightChannel;
         delete[] barPeaksBothChannels;
         delete[] barPeaksLeftChannel;
         delete[] barPeaksRightChannel;
      }
   }
}
