/**
 *  GraphTFT plugin for the Video Disk Recorder 
 * 
 *  display.h - A plugin for the Video Disk Recorder
 *
 *  (c) 2004-2006 Lars Tegeler, Sascha Volkenandt, Jrg Wendel
 *
 * This code is distributed under the terms and conditions of the
 * GNU GENERAL PUBLIC LICENSE. See the file COPYING for details.
 *
 * $Id: display.h,v 1.2 2006/11/06 10:29:23 root Exp $
 *
 **/

#ifndef __GTFT_DISPLAY_H
#define __GTFT_DISPLAY_H

//***************************************************************************
// Includes
//***************************************************************************

#include <map>
#include <string>
#include <vector>

#include <vdr/thread.h>
#include <vdr/status.h>
#include <vdr/tools.h>

#include <common.h>
#include <comthread.h>
#include <renderer.h>
#include <theme.h>
#include <transfer.h>

using std::map;
using std::string;
using std::vector;

//***************************************************************************
// Class cGraphTFTService
//***************************************************************************

class cGraphTFTService : public cThemeService
{
   public:

      enum ChannelType
      {
         ctRadio = 0,
         ctTv
      };

      enum NormalMode
      {
         nmStandard = 0,
         nmClock,
         nmLogo,
         nmSysinfo,
         nmSoccer
      };

      static const char* normalModes[];
};

//***************************************************************************
// Class Span Service
//***************************************************************************

#define SPAN_PROVIDER_CHECK_ID  "Span-ProviderCheck-v1.0"
#define SPAN_CLIENT_CHECK_ID 	  "Span-ClientCheck-v1.0"
#define SPAN_SET_PCM_DATA_ID 	  "Span-SetPcmData-v1.0"
#define SPAN_GET_BAR_HEIGHTS_ID "Span-GetBarHeights-v1.0"

//***************************************************************************
// Calss cSpanService
//***************************************************************************

class cSpanService
{
   public:
      
      // Span requests to collect possible providers / clients

      struct Span_Provider_Check_1_0 
      {
         bool* isActive;
         bool* isRunning;
      };

      struct Span_Client_Check_1_0 
      {
         bool* isActive;
         bool* isRunning;
      };

      // Span data

      struct Span_SetPcmData_1_0 
      {
         unsigned int length;	// the length of the PCM-data
         int* data;				// the PCM-Data as 32-bit int, however only the lower 16-bit are used
                              // and you have to take care to hand in such data!
      };

      struct Span_GetBarHeights_v1_0 
      {
         unsigned int bands;						    // number of bands to compute
         unsigned int* barHeights;				    // the heights of the bars of the two channels combined
         unsigned int* barHeightsLeftChannel;	 // the heights of the bars of the left channel
         unsigned int* barHeightsRightChannel;	 // the heights of the bars of the right channel
         unsigned int* volumeLeftChannel;		    // the volume of the left channels
         unsigned int* volumeRightChannel;		 // the volume of the right channels
         unsigned int* volumeBothChannels;		 // the combined volume of the two channels
         const char* name;					       	 // name of the plugin that wants to get the data
                                                 // (must be unique for each client!)
         unsigned int falloff;                   // bar falloff value
         unsigned int* barPeaksBothChannels;     // bar peaks of the two channels combined
         unsigned int* barPeaksLeftChannel;      // bar peaks of the left channel
         unsigned int* barPeaksRightChannel;     // bar peaks of the right channel
      };
};

//***************************************************************************
// Class GraphTFT Display
//***************************************************************************

class cGraphTFTDisplay : public cStatus, cThread, cGraphTFTService, cSpanService
{
   public:

      enum { MaxTabs = 6 };
      
      enum DisplayMode 
      {
         ModeUnknown      = 0x00,

         ModeMask         = 0xF0,
         DisplayMask      = 0x0F,

         ModeNormal       = 0x10,
         ModeReplay       = 0x20,
         ModeMenu         = 0x30,
         ModeCallibration = 0x40,

         NormalView       = 0x11,

         ReplayNormal     = 0x21,
         ReplayMP3        = 0x22,
         ReplayDVD        = 0x23,

         MenuDefault      = 0x31,
         MenuMain         = 0x32,
         MenuSchedule     = 0x33,
         MenuChannels     = 0x34,
         MenuTimers       = 0x35,
         MenuRecordings   = 0x36,
         MenuSetup        = 0x37,
         MenuCommands     = 0x38,
         MenuWhatsOnElse  = 0x39,
         MenuWhatsOnNow   = 0x3a,
         MenuWhatsOnNext  = 0x3b,
         MenuSetupPage    = 0x3c,

         Calibration      = 0x41
      };

      int isModeNormal(DisplayMode mode)       { return (mode & ModeMask) == ModeNormal; }
      int isModeReplay(DisplayMode mode)       { return (mode & ModeMask) == ModeReplay; }
      int isModeMenu(DisplayMode mode)         { return (mode & ModeMask) == ModeMenu; }
      int isModeCallibration(DisplayMode mode) { return (mode & ModeMask) == ModeCallibration; }

      void setMode(DisplayMode mode, int force = false)
      { 
         if (_mode == Calibration && !force)
            return;
         
         if (_mode != mode)
         {
            _mode = mode;

            d(1, "Mode is set to (0x%X)", _mode);
         }
      }
         
   private:

      cTimeMs timeMs;
      int tabs[MaxTabs+1];
      int charInTabs[MaxTabs+1];

      // some structures and types

      typedef vector <string> Strings;

      struct RecordingInfo 
      {
         string name;
         const cDevice *device;
      };

      typedef vector <RecordingInfo> Recordings;
   
      struct ReplayInfo 
      {
         DisplayMode lastMode;
         string name;
         string folder;
         cControl* control;
      };

      struct CalibrationPos
      {
         int x;
         int y;
      };

      enum MenuItemType
      {
         itNormal,
         itPartingLine
      };

      struct MenuItem 
      {
         string text;
         string tabs[MaxTabs];
         int tabCount;
         MenuItemType type;
      };
   
      struct MenuInfo 
      {
         DisplayMode lastMode;
         int currentPos;
         int currentPosLast;
         int topPos;
         int lineHeight;
         string title;
         string current;
         string buttons[4];
         string text;
         int textScroll;
         vector<MenuItem> items;
      };

      // renderer

      Renderer* _render;
      ComThread* comThread;

      // update thread

      bool _active;
      bool _marquee_active;
      bool _marquee_left;
      int _marquee_idx;
      int _marquee_count;
      int _marquee_strip;
      int _updateIn;
      int _Dumped;
      int _recordingIn;
      int _volumeIn;
      int _messageIn;
      int _delayIn;
      cMutex _mutex;
      cCondVar _doUpdate;

      // animation stuff

      bool _animation;
      int _displayFrame;
      int _frameIn;

      // display contents

      cThemeSection* currentSection;
      ChannelType channelType;
      DisplayMode _mode;
      NormalMode normalMode;       // mode used for NormalTV and NormalRadio
      bool _mute;
      int _channel;
      int _volume;
      int _showRecording;
      time_t _presentTime;
      time_t _followingTime;
      string _channelText;
      string _channelInfo;
      string _presentTitle;
      string _presentSubtitle;
      string _followingTitle;
      string _followingSubtitle;
      string _message;
      Recordings _recordings;
      ReplayInfo _replay;
      MenuInfo _menu;
      CalibrationPos calibrationPos;

      // init pbp stuff

      bool _pbpActive;
      cDevice* _dev;
      const cChannel* _chan;
      cGraphTFTTransfer* _transfer;
      int menuCount;

   protected:

      // update thread

      virtual void Action();

      // helper functions

      bool StoppedTimer(const char *Name);
      void UpdateProgramme();

      // status interface

      virtual void ChannelSwitch(const cDevice *Device, int ChannelNumber);
      virtual void OsdChannel(const char *Text);
      virtual void OsdProgramme(time_t PresentTime, const char *PresentTitle, const char *PresentSubtitle, time_t FollowingTime, const char *FollowingTitle, const char *FollowingSubtitle);
      virtual void SetVolume(int Volume, bool Absolute);
      virtual void Replaying(const cControl *Control, const char *Name, const char *FileName, bool On);
      virtual void Recording(const cDevice *Device, const char *Name, const char *FileName, bool On);
      virtual void OsdStatusMessage(const char *Message);
      virtual void OsdTitle(const char* Title);
      virtual void OsdItem(const char* Text, int Index);
      virtual void OsdCurrentItem(const char *Text);
      virtual void OsdClear(void);
      virtual void OsdHelpKeys(const char* Red, const char* Green, const char* Yellow, const char* Blue);
      virtual void OsdTextItem(const char* Text, bool Scroll);
      virtual void OsdMenuDisplay(eOsdMenuKind kind);
      virtual void OsdMenuDestroy();

   public:

      cGraphTFTDisplay();
      ~cGraphTFTDisplay();

      int Init(const char* dev, const char* cfgDir, int port);
      void setupChanged();
      void setCalibrate(int active);
   
      // thread stuff

      bool Start()        { return cThread::Start(); }
      void Stop();
      bool Active() const { return _active; }
      cMutex* getMutex()  { return &_mutex; }

      void setDefaultMode(int defMode) { normalMode = (NormalMode)defMode; }

      // event
      
      void mouseEvent(int x, int y, int button, int flag = ComThread::efNone);

      // PbP stuff

      void Pbp();
      void PbpRestart();
      void PbpSwitch(bool Back);
      void PbpUpDown(bool Up);
      bool PbpIsRunning();
      void PbpStart();
      void PbpStop();
      void PbpInit();

      // display stuff

      void Display(int sectionId);
      void DisplayImage(int x, int y, int width, int height, int transparent, string image, int fit = false);
      void DisplayImageFile(int x, int y, int width, int height, int transparent, string file, string image);
      void DisplayImageCover(int x, int y, int width, int height, int transparent, string file);
      void DisplayDate(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayDateOnly(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayTimeOnly(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayText(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines, string text, bool marquee = false);
      void DisplayChannelName(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayChannelNameOnly(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayChannelNumber(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayRectangle(int x, int y, int width, int height, int transparent, int red, int green, int blue);
      void DisplayLogo(int x, int y, int width, int height, int transparent, string image ,string type, int animated, int delay, int count);
      void DisplayPresentTitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayPresentSubtitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayPresentTime(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayFollowingTitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayFollowingSubtitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayFollowingTime(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayMessage(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayRecording(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayReplay(int Progress, int Current, int Total); 
      void DisplayMenuTitle(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines );
      void DisplayMenuText(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines );
      void DisplayMenuTextBackground(int x, int y, int width, int height, int transparent, string image);
      void DisplayMenuButton(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines, int buttonNumber);
      void DisplayMenuButtonBackground(int x, int y, int width, int height, int transparent, string image, string imageOff, int buttonNumber);
      void DisplayVolumebar( int x, int y ,int width, int height, int red, int green, int blue, int transparent  );
      void DisplaySymbols(int x, int y, int width, int height, int transparent, string path, string path2, int symbol);
      void DisplayRecordingSymbol(int x, int y, int width, int height, int transparent, string path, string path2);
      void DisplayReplayName(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayReplayFolder(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayReplayProcessBar(int x, int y ,int width, int height, int red, int green, int blue, int transparent, int bg_x, int bg_y ,int bg_width, int bg_height, int bg_red, int bg_green, int bg_blue, int bg_transparent);
      void DisplayReplayCurrent(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayReplayTotal(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayReplayRemaining(int x, int y, int width, int height,int red, int green, int blue, int transparent, string font, int size, int align, int lines);
      void DisplayMessageBackground(int x, int y, int width, int height, int transparent, string image);
      void DisplayVolumeBackground(int x, int y, int width, int height, int transparent, string image);
      void DisplayVolumeMuteSymbole(int x, int y, int width, int height, int transparent, string image, string image2);
      void DisplayShow();
      void DisplayClear();

      void DisplayTimebar(cGraphTFTThemeItem* item);
      void DisplayMenuDefault(cGraphTFTThemeItem* item, cGraphTFTThemeItem* selectedItem);
      void DisplayMenuColumn(cGraphTFTThemeItem* item, cGraphTFTThemeItem* selectedItem);
      void DisplaySpectrumAnalyzer(cGraphTFTThemeItem* item);
      void DisplayPartingLine(cGraphTFTThemeItem* p, string text, int y, int height);
      void DisplayProgressBar(cGraphTFTThemeItem* p, float current, float total, int y = -1, int height = -1);
      void DisplayCalibrationCursor(cGraphTFTThemeItem* p);
      void DisplayDateTimeFormat(cGraphTFTThemeItem* p);

      // Tabs

      int Tab(int n) { return (n >= 0 && n <= MaxTabs) ? tabs[n] : 0; }
      const char* GetTabbedText(const char* s, int Tab);
      virtual void SetTabs(int Tab1, int Tab2 = 0, int Tab3 = 0, int Tab4 = 0, int Tab5 = 0);

};

//***************************************************************************
#endif // __GTFT_DISPLAY_H
